#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import datetime

from calcifer import __version__

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

project = 'calcifer'
copyright = '{0.year}, Dramafever'.format(datetime.datetime.now())
version = release = __version__

needs_sphinx = '1.1'
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.graphviz',
    'sphinx.ext.autosummary',
    'sphinx.ext.mathjax',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
]

templates_path = []
source_suffix = '.rst'
source_encoding = 'utf-8-sig'
master_doc = 'index'
pygments_style = 'sphinx'
html_static_path = []
exclude_patterns = []
graphviz_output_format = 'svg'

if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme  # pylint: disable=C0413
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

intersphinx_mapping = {
    'python': ('https://docs.python.org/', None),
}



from abc import ABCMeta
import collections
import copy
import re
from six import string_types


def get_call_repr(func_name, *args, **kwargs):
    args_expr = ", ".join([repr(arg) for arg in args])
    kwargs_expr = ", ".join([
        "{k}={v}".format(k=k, v=repr(v))
        for k, v in kwargs.items()
    ])
    call_expr = "{name}(".format(name=func_name)
    call_expr += args_expr
    if args_expr and kwargs_expr:
        call_expr += ", "
    call_expr += kwargs_expr
    call_expr += ")"
    return call_expr


class Node:
    __metaclass__ = ABCMeta

    def __new__(cls, *operands):
        return super(Node, cls).__new__(cls)

    def __deepcopy__(self, memo):
        return copy.copy(self)


def flatten(l):
    for el in l:
        if isinstance(el, collections.Iterable) and not isinstance(el, string_types):
            for sub in flatten(el):
                yield sub
        else:
            yield el


class Binding(Node):
    def __new__(cls, *operands):
        operands = flatten([
            operand if not isinstance(operand, cls) else operand.operands
            for operand in operands
        ])
        return super(Binding, cls).__new__(cls, operands)

    def __init__(self, *operands):
        self.operands = operands

    def __repr__(self):
        return " >> ".join([repr(operand) for operand in self.operands])


class PolicyRuleFunc(Node):
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name


class PolicyRuleFuncCall(Node):
    def __init__(self, func, args, kwargs, result=None):
        self.func = func
        self.args = [
            getattr(arg, 'ast', arg)
            for arg in args
        ]
        self.kwargs = {
            k: getattr(v, 'ast', v)
            for k, v in kwargs.items()
        }
        self.result = result

    def with_result(self, result):
        return PolicyRuleFuncCall(
            self.func,
            self.args,
            self.kwargs,
            result
        )

    def __repr__(self):
        if isinstance(self.func, Node):
            func_name = repr(self.func)
        else:
            func_name = self.func

        identifier = re.compile(r"^[^\d\W]\w*\Z", re.UNICODE)

        if not re.match(identifier, func_name):
            func_name = "({})".format(func_name)

        return get_call_repr(func_name, *self.args, **self.kwargs)



# -*- coding: utf-8 -*-
"""
calcifer top-level module

The purpose of this module is to provide runtime validation and template
generation for commands.

This module provides low-level operators to describe the non-deterministic
manipulation of a "Policy Partial" data structure to be used in validation and
template generation. The operators are designed to provide flexible
tooling for the creation of high-level policy rules.
"""
from calcifer.contexts import Context
from calcifer.operators import (
    attempt,
    append_value,
    catch_attempt,
    check,
    children,
    collect,
    define_as,
    each,
    fail,
    forbid_value,
    get_node,
    get_value,
    match,
    permit_values,
    policies,
    pop_value,
    pop_context,
    push_context,
    regarding,
    require_value,
    scope,
    select,
    set_value,
    trace,
    unit,
    unit_value,
    unless_errors,
    wrap_context,
)
from calcifer.partial import Partial
from calcifer.monads import (
    PolicyRule, PolicyRuleFunc, policy_rule, policy_rule_func,
)
from calcifer.policy import BasePolicy
from calcifer.policy import DefaultPolicy as Policy

from calcifer._version import __version__



import random
import logging

import unittest
from unittest import TestCase

from calcifer.utils import run_policy

from calcifer.contexts.base import Incomplete
from calcifer.contexts import (
    Context
)

from calcifer import (
    regarding, set_value, unit,

    asts, PolicyRule
)

logger = logging.getLogger(__name__)


class ContextTestCase(TestCase):
    def test_append_policy(self):
        ctx = Context()
        policy = regarding('/foo', set_value(5))
        ctx.append(policy)
        policy = ctx.finalize()

        result = run_policy(policy)

        self.assertEqual(result['foo'], 5)

    def test_append_function(self):
        ctx = Context()
        value = unit(5)

        def with_value(value):
            return regarding(
                '/foo',
                set_value(value)
            )

        ctx.append(with_value, value)
        policy = ctx.finalize()

        result = run_policy(policy)
        self.assertEqual(result['foo'], 5)

    def test_subctx_policy(self):
        ctx = Context()

        subctx = ctx.subctx(
            lambda policy_rules: regarding('/foo', *policy_rules)
        )

        subctx.append(
            set_value(5)
        )

        result = run_policy(ctx.finalize())
        self.assertEqual(result['foo'], 5)

    def test_subctx_noop_policy(self):
        ctx = Context()

        foo_ctx_value = ctx.subctx(
            lambda policy_rules: regarding('/foo', *policy_rules)
        )

        def with_foo(foo_value):
            return regarding('/bar', set_value(foo_value))

        ctx.append(with_foo, foo_ctx_value)

        # foo is only used as a value - never being applied policy
        # rules itself.
        # ctx therefore should only have 1 policy, the `with_foo`
        # function above that just sets {foo: *, bar: foo}
        items, error_handler = ctx.get_finalized_items()
        self.assertEqual(len(items), 1)

        result = run_policy(ctx.finalize(), {"foo": "zebra"})
        self.assertEqual(result['bar'], "zebra")

    def test_subclass_helpers(self):
        class ResourceContext(Context):
            @property
            def resource(self):
                return self.scope_subctx("/resource", "resource")

            def require_resource(self):
                require_ctx = ctx.resource.require()
                error_ctx = require_ctx.error_ctx()

                error_ctx.select("message").set_value(
                    "Resource is required"
                )
                error_ctx.select("code").set_value("MISSING_REQUIRED_RESOURCE")
                return require_ctx

        ctx = ResourceContext()
        ctx.require_resource()

        result = run_policy(ctx.finalize(), {})

        error = result["errors"][0]

        self.assertEqual(error["scope"], "/resource")
        self.assertEqual(error["value"], None)
        self.assertEqual(len(error["context"]), 2)

        expected_context_names = ["resource", "require"]
        actual_context_names = [frame.name for frame in error["context"]]
        self.assertEqual(expected_context_names, actual_context_names)

        policy_asts = [frame.policy_ast for frame in error["context"]]

        for ast in policy_asts:
            self.assertIsInstance(ast, asts.Node)

    def test_whitelist_values(self):
        ctx = Context()
        ctx.select("/client").whitelist_values(
            ["ios", "android"]
        )

        obj = {
            "client": "windows"
        }
        result = run_policy(ctx.finalize(), obj)

        error = result["errors"][0]

        error_context_names = [frame.name for frame in error['context']]
        self.assertIn("whitelist_values", error_context_names)

    def test_each(self):
        ctx = Context(name="root")
        ref_obj = {
            "a": 1,
            "c": 3,
            "d": 4,
        }
        eachctx = ctx.select("/dict").each(ref=ref_obj)
        ref_value = eachctx.value

        eachctx.require(ref_value)

        result = run_policy(ctx.finalize(), {
            "dict": {"a": 0, "b": 0, "c": 0, "d": 0}
        })

        self.assertNotIn("data", result)
        self.assertIn("errors", result)

        errors = result["errors"]

        self.assertEqual(len(errors), 1)
        error = errors[0]
        self.assertEqual(error["scope"], "/dict/b")
        self.assertEqual(error["message"], "Value is required.")

    def test_finalize(self):
        ctx = Context(name="root")
        a = ctx.select("/a")
        b = a.select("/b")
        b.set_value(a.value)

        incomplete = b.finalize()
        self.assertIsInstance(incomplete, Incomplete)
        self.assertIn(a.value, incomplete.missing)

        completed = incomplete.complete({a.value: 5})
        self.assertIsInstance(completed, PolicyRule)
        result = run_policy(completed)
        self.assertEqual(result['b'], 5)

    def test_finalize_own_ctx_value(self):
        ctx = Context(name="root")
        a = ctx.select("/a")
        b = a.select("/b")
        b.set_value(b.value)

        completed = b.finalize()
        self.assertIsInstance(completed, PolicyRule)

        result = run_policy(completed, {"b": 5})
        self.assertEqual(result['b'], 5)

        completed = ctx.finalize()
        self.assertIsInstance(completed, PolicyRule)

        result = run_policy(completed, {"b": 5})
        self.assertEqual(result['b'], 5)

    def test_finalize_two_ctx_values(self):
        ctx = Context(name="root")
        a = ctx.select("/a")
        b = a.select("/b")
        c = b.select("/c")

        def func(b, a):
            return set_value(a + b)

        c.append(func, a.value, b.value)

        incomplete = c.finalize()
        self.assertIsInstance(incomplete, Incomplete)
        self.assertIn(a.value, incomplete.missing)
        self.assertIn(b.value, incomplete.missing)

        completed = incomplete.complete({a.value: 5, b.value: 2})
        self.assertIsInstance(completed, PolicyRule)

        result = run_policy(completed)
        self.assertEqual(result['c'], 7)

        # now instead of starting with `c` and supplying 2 values,
        # let's start with `b` and only have to supply one value.
        incomplete = b.finalize()
        self.assertIsInstance(incomplete, Incomplete)
        self.assertIn(a.value, incomplete.missing)
        self.assertNotIn(b.value, incomplete.missing)

        completed = incomplete.complete({a.value: 5})
        self.assertIsInstance(completed, PolicyRule)

        result = run_policy(completed, {"b": 2})
        self.assertEqual(result['c'], 7)

    def test_incomplete_hookup(self):
        ctx = Context(name="root")
        a = ctx.select("/a")
        b = a.select("/b")
        b.set_value(a.value)

        complete = a.finalize()
        self.assertIsInstance(complete, PolicyRule)
        result = run_policy(complete, {"a": 5})
        self.assertEqual(result['b'], 5)

        complete = ctx.finalize()
        self.assertIsInstance(complete, PolicyRule)
        result = run_policy(complete, {"a": 5})
        self.assertEqual(result['b'], 5)

    def test_incomplete_hookup_skipping(self):
        ctx = Context(name="root")
        a = ctx.select("/a")
        b = a.select("/b")
        c = b.select("/c")
        c.set_value(a.value)

        complete = a.finalize()
        self.assertIsInstance(complete, PolicyRule)
        result = run_policy(complete, {"a": 5})
        self.assertEqual(result['c'], 5)

        complete = ctx.finalize()
        self.assertIsInstance(complete, PolicyRule)
        result = run_policy(complete, {"a": 5})
        self.assertEqual(result['c'], 5)

    def test_nesting(self):
        ctx = Context(name="root")

        a = ctx.select("/a")
        b = a.select("/b")

        def f_a(a):
            return regarding("/results/f_a", set_value(a))

        def f_b(b):
            return regarding("/results/f_b", set_value(b))

        b.append(f_a, a.value)
        b.append(f_b, b.value)

        c = b.select("/c")

        def f_ab(a, b):
            return regarding("/results/f_ab", set_value(a + b))

        def f_ac(a, c):
            return regarding("/results/f_ac", set_value(a + c))

        c.append(f_ab, a.value, b.value)
        c.append(f_ac, a.value, c.value)

        result = run_policy(ctx.finalize(), {
            "a": 1,
            "b": 2,
            "c": 4,
        })

        a = result['a']
        b = result['b']
        c = result['c']

        f_a = result['results']['f_a']
        f_b = result['results']['f_b']
        f_ab = result['results']['f_ab']
        f_ac = result['results']['f_ac']

        self.assertEqual(f_a, a)
        self.assertEqual(f_b, b)
        self.assertEqual(f_ab, a + b)
        self.assertEqual(f_ac, a + c)

    def test_append_mixed(self):
        ctx = Context(name="root")

        a = ctx.select("/a")
        b = a.select("/b")

        def with_values(a, b):
            return set_value(a + b)

        b.append(with_values, a.value, unit(2))

        result = run_policy(ctx.finalize(), {"a": 5})
        self.assertEqual(result['b'], 7)

    def test_apply(self):
        ctx = Context(name="root")
        a = ctx.select("/a")
        b = a.select("/b")

        def increment(x):
            return x + 1

        # the value only exists inside the context of the function
        # application
        applied_ctx = b.apply(increment, a.value)
        applied_ctx.set_value(applied_ctx.value)

        result = run_policy(ctx.finalize(), {"a": 5})
        self.assertEqual(result['b'], 6)

        # and just for calisthenics, do it in a different order
        ctx = Context(name="root")
        a = ctx.select("/a")
        applied_ctx = a.apply(increment, a.value)
        b = applied_ctx.select("/b")
        b.set_value(applied_ctx.value)

        result = run_policy(ctx.finalize(), {"a": -3})
        self.assertEqual(result['b'], -2)

    def test_apply_alchemy(self):
        # for our test today, we will be doing some basic alchemy
        inventory = [
            "aqua fortis",
            "glauber's salt",
            "lunar caustic",
            "mosaic gold",
            "plumbago",
            "salt",
            "tin salt",
            "butter of tin",
            "stibnite",
            "naples yellow",
        ]

        # backstory:
        # ~~~~~~~~~~
        #
        # falling asleep last night, you finally figured out how to complete
        # your life's work: discovering the elusive *elixir of life*!
        #
        # and it's only two ingredients! and you have them on hand!
        #
        # ...
        #
        # unfortunately this morning you can't remember which two
        # ingredients it was.
        #
        # you'll know it once you've gotten it, just have to try out
        # all possible mixtures. (should be safe enough, right?)

        forgotten_elixir_of_life = set(random.sample(inventory, 2))

        discoveries_today = set(["frantic worry", "breakfast"])

        # ok time to go do some arbitrary alchemy!
        #
        # game plan:
        alchemy_ctx = Context()

        # you'll grab one ingredient,
        selected_first_ctx = alchemy_ctx.select("/inventory").each()
        first_substance = selected_first_ctx.value

        # and another,
        selected_second_ctx = selected_first_ctx.select("/inventory").each()
        second_substance = selected_second_ctx.value

        # take them to your advanced scientific mixing equipment,
        workstation_ctx = selected_second_ctx.select("/workstation")

        # (btw this is your advanced scientific procedure that you are
        # 100% certain will tell you what some mixture is)
        def mix(first, second):
            """takes two ingredients and returns the resulting substance"""
            if set([first, second]) == forgotten_elixir_of_life:
                return "elixir of life"
            return "some kind of brown goo"

        # then you'll mix your ingredients...
        mixed_ctx = workstation_ctx.apply(
            mix,
            first_substance, second_substance
        )
        resulting_mixture = mixed_ctx.value

        # ... and! in today's modern age, scientists now know to record their
        # results!
        mixed_ctx.select("/discoveries").append_value(resulting_mixture)

        # got it? good!
        result = run_policy(
            alchemy_ctx.finalize(),
            {"inventory": inventory, "discoveries": discoveries_today}
        )

        # in a flurry of excitement, i bet you didn't even stop to
        # look at your discoveries as you made them!
        #
        # well, let's see...

        self.assertIn("elixir of life", result["discoveries"])

    def test_apply_dangerous_alchemy(self):
        # nice job! and you even finished in time to go foraging for
        # more ingredients!
        inventory = [
            "aqua fortis",
            "glauber's salt",
            "lunar caustic",
            "mosaic gold",
            "plumbago",
            "salt",
            "tin salt",
            "butter of tin",
            "stibnite",
            "naples yellow",

            # nice find
            "anti-plumbago"
        ]

        # but unfortunately, it's the next day, and the same thing
        # has happened to you! except this time it was for your
        # other life's goal: discover the ~elixir of discord~!
        #
        # well, since it was so easy...

        whatever_concoction = set(['some ingredients'])

        discoveries_today = set([])
        should_be_fine = 'overconfidence' not in discoveries_today
        assert should_be_fine

        # doing alchemy la la la
        alchemy_ctx = Context()

        # grabbin' things off shelves
        selected_first_ctx = alchemy_ctx.select("/inventory").each()
        first_substance = selected_first_ctx.value

        selected_second_ctx = selected_first_ctx.select("/inventory").each()
        second_substance = selected_second_ctx.value

        # got our ingredients
        got_ingredients_ctx = selected_second_ctx

        workstation_ctx = got_ingredients_ctx.select("/workstation")

        # mixin' - don't stop to think
        def mix(first, second):
            mixture = set([first, second])
            if mixture == whatever_concoction:
                return 'missing elixir'
            if mixture == set(['plumbago', 'anti-plumbago']):
                return 'concentrated danger'
            return 'more brown goo'

        mixed_ctx = workstation_ctx.apply(
            mix,
            first_substance, second_substance
        )
        resulting_mixture = mixed_ctx.value

        mixed_ctx.select("/discoveries").append_value(resulting_mixture)

        # wait wait wait!!
        def danger(mixture):
            if mixture == 'concentrated danger':
                return True
            return False

        # we can't have that.
        danger_ctx = mixed_ctx.check(
            danger,
            resulting_mixture
        )
        danger_ctx.forbid()

        # moral:
        #
        # a strong understanding of policies and processes facilitates a
        # hazard-free lab environment.
        result = run_policy(
            alchemy_ctx.finalize(),
            {"inventory": inventory, "discoveries": discoveries_today}
        )

        self.assertIn("errors", result)
        self.assertTrue(len(result['errors']))


if __name__ == '__main__':
    unittest.main()



import logging

from calcifer.operators import (
    set_value, permit_values, require_value, append_value,
    forbid_value, children, each, collect, unless_errors,
)
from calcifer.contexts.policies import (
    add_error
)
from calcifer.contexts.base import BaseContext

logger = logging.getLogger(__name__)


class Context(BaseContext):
    """
    `Context` provides a high-level interface for building policies.

    Policies are built by performing stateful operations on Context objects.

    Method calls/property retrievals modify the Context to potentially
    include additional policy rules, as appropriate.

    Implementation details can be found in BaseContext
    """
    def error_ctx(self):
        """
        Retrieves, possibly creating, a specialized error handler
        policy for the Context
        """
        if not self.error_handler:
            error_handler_ctx = self.__class__(name="error_handler")
            self.error_handler = error_handler_ctx
        return self.error_handler

    def require(self, *args):
        """
        Requires that a value is defined and truthy.

        :param value: if not provided, uses value for current node
        """
        if args:
            value = args[0]
        else:
            value = self.value

        subctx = self.named_subctx("require")

        error_ctx = subctx.error_ctx()
        error_ctx.select("code").set_value(
            "MISSING_REQUIRED_VALUE"
        )
        error_ctx.select("message").set_value(
            "Value is required."
        )

        subctx.append(require_value, value).or_error()
        return subctx

    def forbid(self, *args):
        """
        Opposite of ``require()`` - errors when value is defined
        """
        if args:
            value = args[0]
        else:
            value = self.value
        subctx = self.named_subctx("forbid")
        subctx.append(forbid_value, value).or_error()
        return subctx

    def set_value(self, value):
        """
        Sets the value for the current node
        """
        self.append(set_value, value)
        return self

    def append_value(self, value):
        """
        Appends value to the current node, assuming the node
        to be a list if not defined
        """
        self.append(append_value, value)
        return self

    def whitelist_values(self, values):
        """
        Forks computation, erring if value is provided already and
        does not match
        """
        subctx = self.named_subctx("whitelist_values")
        subctx.append(permit_values, values).or_error()

        error_ctx = subctx.error_ctx()
        error_ctx.select("code").set_value("INVALID_VALUE_SELECTION")
        error_ctx.select("values").set_value(values)

        return subctx

    def fail_early(self):
        """
        Returns a new context that checks node "/errors" and short-circuits
        if any errors exist.
        """
        return self.subctx(
            lambda policy_rules: unless_errors(*policy_rules)
        )

    def err(self):
        """
        Trigger error handling
        """
        return self.fail().or_error()

    def add_error(self):
        """
        Create a blank error
        """
        self.append(add_error, {})

    @property
    def last_error(self):
        """
        Returns the context selecting the most recently defined error
        """
        errors_ctx = self.select("/errors").children()
        idx_ctx = errors_ctx.apply(lambda es: es[-1], errors_ctx.value)
        last_error_ctx = idx_ctx.select(idx_ctx.value)
        return last_error_ctx

    def or_error(self):
        """
        If context fails, inject error instead.
        Error has the following properties:

        value
            Value found at node
        scope
            The current scope at the time of error

        context
            The contextual traceback
        """

        catch_ctx = self.or_catch()

        subctx = catch_ctx.subctx()
        subctx.add_error()
        # prepare error

        # include provided value in error
        provided_value_ctx = subctx.apply(
            lambda true_trace_obj: true_trace_obj['value'],
            catch_ctx.value
        )
        provided_value_ctx.last_error.select("value").set_value(provided_value_ctx.value)

        # include scope
        scope_ctx = subctx.apply(
            lambda true_trace_obj: true_trace_obj['scope'],
            catch_ctx.value
        )
        scope_ctx.last_error.select("scope").set_value(scope_ctx.value)

        # include context frameset itself
        ctxes_ctx = subctx.apply(
            lambda true_trace_obj: true_trace_obj['context'],
            catch_ctx.value
        )
        ctxes_ctx.last_error.select("context").set_value(ctxes_ctx.value)

        # for each frame in the context, allow the ctx frame to specify
        # an error handler policy rule that will be:

        # for each ctx frame,
        ctx_list_ctx = subctx.last_error.select("context")

        def for_ctx_list(ctx_list):
            return ctx_list
        ctx_list_ctx.apply(for_ctx_list, ctx_list_ctx.value)

        frame_ctx = ctx_list_ctx.last_error.select("context").each()
        frame_ctx.select("context").apply
        ctx_frame = frame_ctx.value

        # checking for error_handler,
        error_handler_ctx = frame_ctx.check(
            lambda true_frame: true_frame.error_handler,
            ctx_frame
        )

        # append the error handler as a policy rule
        error_handler_ctx.last_error.append(
            error_handler_ctx.value
        )

        return self

    def children(self):
        """
        Return the context with the list of scopes that are direct children
        of the current node
        """
        children_ctx = self.subctx(
            lambda policy_rules: (
                children() >> collect(*policy_rules)
            )
        )
        return children_ctx

    def each(self, **kwargs):
        """
        Create and return a context that operates on each child of the
        current node.

        :kwarg ref: An injectable reference object that has matching children
            nodes (same structure dict or list)
        """
        def with_policy_rules(policy_rules):
            def with_true_children(true_children):
                return each(
                    *policy_rules, **kwargs
                )(true_children)
            return with_true_children

        subctx = self.subctx()
        eachctx = subctx.named_subctx(
            "each",
            with_policy_rules,
            subctx.children()
        )
        return eachctx



"""
Premium Command Policy StateT Operators.

These are provided as building blocks for specifying Premium Command Policies
for the purposes of template generation and command validation.
"""
import copy
import logging
from pymonad import List

from calcifer.tree import PolicyNode
from calcifer.monads import (
    policy_rule_funcM as policy_rule_func,
    get_call_repr, PolicyRule
)

logger = logging.getLogger(__name__)


#
# Partial Operators
#

def make_unit(m):
    @policy_rule_func(m)
    def unit(value):
        """
        Returns a value inside the monad

        :param value: the value returned inside the PolicyRule monad
        """
        def for_partial(partial):
            return m.unit((value, partial))
        return for_partial
    return unit


unit = make_unit(List)


def make_unit_value(m):
    @policy_rule_func(m)
    def unit_value(node):
        """
        Given a node (often returned as monadic result), return
        the value for the node.

        :param node: the node whose value is to be returned inside the
            PolicyRule monad
        :type node: Node _v_
        :returns:    PolicyRule _v_
        """
        def for_partial(partial):
            if hasattr(node, 'value'):
                return m.unit((node.value, partial))
            return m.unit((None, partial))
        return for_partial
    return unit_value


unit_value = make_unit_value(List)


def make_set_value(m):
    @policy_rule_func(m)
    def set_value(value):
        """
        Sets the value for the currently scoped policy node. Overwrites
        the node with a LeafPolicyNode

        :param value: new value
        :type value: *v*
        :returns: PolicyRule *v*
        """
        def for_partial(partial):
            return m.unit(partial.set_value(value))
        return for_partial
    return set_value


set_value = make_set_value(List)


def make_select(m):
    @policy_rule_func(m)
    def select(scope, set_path=False):
        """
        Retrieves the policy node at a given selector and optionally
        sets the scope to that selector. Recursively defines UnknownPolicyNodes
        in the partial.

        :param scope: Scope to select

        :keyword set_path: Sets the scope
        :type set_path: bool

        :type scope: json pointer string

        :returns: PolicyRule (Node *v*)
        """
        def for_partial(partial):
            return m.unit(partial.select(scope, set_path=set_path))
        return for_partial
    return select


select = make_select(List)


def make_scope(m):
    @policy_rule_func(m)
    def scope():
        """
        Returns the current scope for the partial

        :returns: PolicyRule *string json pointer*
        """
        def for_partial(partial):
            return m.unit((partial.scope, partial))
        return for_partial
    return scope


scope = make_scope(List)


def make_get_node(m):
    @policy_rule_func(m)
    def get_node():
        """
        Retrieves the node at the current scope

        :returns: PolicyRule (Node *v*)
        """
        def for_partial(partial):
            return m.unit(partial.select(""))
        return for_partial
    return get_node


get_node = make_get_node(List)


def make_children(m):
    @policy_rule_func(m)
    def children():
        """
        For DictPolicyNodes or ListPolicyNodes, returns all scopes that are
        direct children.

        :returns: PolicyRule [scope]
        """
        def for_partial(partial):
            node, _ = partial.select("")
            if not hasattr(node, 'keys'):
                return m.unit(([], partial))

            keys = [str(key) for key in node.keys]
            return m.unit((keys, partial))
        return for_partial
    return children


children = make_children(List)


def make_get_value(m):
    get_node = make_get_node(m)
    unit_value = make_unit_value(m)

    @policy_rule_func(m)
    def get_value():
        """
        Retrieves the value for the node at the current pointer. Equivalent to
        `get_node() >> unit_value`

        :returns: PolicyRule *v*
        """
        return get_node() >> unit_value
    return get_value


get_value = make_get_value(List)


def make_append_value(m):
    get_value = make_get_value(m)
    set_value = make_set_value(m)

    @policy_rule_func(m)
    def append_value(value):
        """
        Gets the value at the current node and appends `value`.
        The current node value should be either a set or a list, or undefined.

        :param value: value to append
        """
        def append_to(collection):
            if isinstance(collection, list):
                return collection + [value]
            elif isinstance(collection, (set, frozenset)):
                return collection | frozenset([value])
            elif collection is None:
                return [value]
            else:
                raise NotImplementedError

        return get_value() >> (
            lambda collection: set_value(append_to(collection))
        )
    return append_value


append_value = make_append_value(List)


def make_pop_value(m):
    get_value = make_get_value(m)
    set_value = make_set_value(m)

    @policy_rule_func(m)
    def pop_value():
        """
        Gets the value at the current node, and pops an element.
        """
        def popped(collection):
            collection = copy.copy(collection)
            if not hasattr(collection, 'pop'):
                raise NotImplementedError

            collection.pop()
            return collection
        return get_value() >> (lambda collection: set_value(popped(collection)))
    return pop_value


pop_value = make_pop_value(List)


def make_define_as(m):
    @policy_rule_func(m)
    def define_as(node):
        """
        Define the node at the current scope

        :param node: Node *v*
        :returns: PolicyRule (Node *v*)
        """
        def for_partial(partial):
            if (
                    not hasattr(node, 'match') and
                    hasattr(node, 'definition')
            ):
                definition = node.definition
            else:
                definition = node

            new_definition, new_partial = partial.define_as(definition)
            if new_definition is None:
                return m.mzero()
            return m.unit((new_definition, new_partial))
        return for_partial
    return define_as


define_as = make_define_as(List)


def make_check(m):
    @policy_rule_func(m)
    def check(func):
        """
        Given a function that takes no arguments, returns a
        policy rule that runs the function and returns the result
        and an unchanged partial
        """
        def for_partial(partial):
            return m.unit((func(), partial))
        return for_partial
    return check


check = make_check(List)


#
# Control Structures
#

def make_collect(m):
    unit = make_unit(m)

    def collect(*rule_funcs):
        """
        Given a list of policy rule functions, returns a single policy rule
        func that accepts some value, provides that to each function,
        resetting the scope each time.
        """
        def for_incoming_value(incoming_value):
            def for_initial_partial(initial_partial):
                initial_scope = initial_partial.scope

                m_results = m.unit((incoming_value, initial_partial))

                def for_rule_func(rule_func):
                    def for_m_result(m_result):
                        _, partial = m_result
                        _, scoped_partial = partial.select(
                            initial_scope, set_path=True
                        )

                        rule = unit(incoming_value) >> rule_func
                        m_results = rule.run(scoped_partial)
                        return m_results
                    return for_m_result

                for rule_func in rule_funcs:
                    for_m_result = for_rule_func(rule_func)
                    m_results = m_results >> for_m_result
                return m_results
            return for_initial_partial

        if not rule_funcs:
            rule_funcs = [unit]

        collect_func_name = get_call_repr('collect', *rule_funcs)
        return policy_rule_func(m, collect_func_name)(for_incoming_value)
    return collect


collect = make_collect(List)


def make_policies(m):
    unit = make_unit(m)

    @policy_rule_func(m)
    def policies(*rule_funcs):
        """
        Given a list of policy rules, returns a single policy rule that
        applies each in turn, keeping scope constant for each. (By resetting
        the path each time)
        """
        def for_initial_partial(initial_partial):
            initial_scope = initial_partial.scope

            m_results = m.unit((None, initial_partial))

            def for_rule_func(rule_func):
                def for_m_result(m_result):
                    _, partial = m_result
                    _, scoped_partial = partial.select(
                        initial_scope, set_path=True
                    )

                    rule = unit(None) >> rule_func
                    m_results = rule.run(scoped_partial)
                    return m_results
                return for_m_result

            for rule_func in rule_funcs:
                for_m_result = for_rule_func(rule_func)
                m_results = m_results >> for_m_result
            return m_results
        return for_initial_partial
    return policies


policies = make_policies(List)


def make_regarding(m):
    policies = make_policies(m)
    select = make_select(m)
    unit_value = make_unit_value(m)

    @policy_rule_func(m)
    def regarding(selector, *rule_funcs):
        """
        Given a selector and a list of functions that generate policy rules,
        returns a single policy rule that, for each rule function:

        1. sets the scope to the selector / retrieves the node there
        3. passes the node to the rule_func to generate a policy rule
        4. applies the policy rule at the new scope

        In addition, regarding checks the current scope and restores it when
        it's done.
        """
        @policy_rule_func(m)
        def regarding_step(rule_func):
            def for_partial(partial):
                original_scope = partial.scope
                node, inner_partial = partial.select(selector, set_path=True)
                value = node.value
                if not value:
                    value = node

                if isinstance(rule_func, PolicyRule):
                    results = rule_func.run(inner_partial)
                else:
                    results = rule_func(value).run(inner_partial)

                def for_result(result):
                    _, partial = result
                    _, rescoped_partial = partial.select(
                        original_scope, set_path=True
                    )
                    return value, rescoped_partial

                return results.fmap(for_result)
            return for_partial

        if rule_funcs:
            op = policies(*[regarding_step(rule_func) for rule_func in rule_funcs])
        else:
            op = select(selector, set_path=False) >> unit_value

        return op

    return regarding


regarding = make_regarding(List)


def make_each(m):
    unit = make_unit(m)

    def each(*rule_funcs, **kwargs):
        """
        `each(rule_func)` is a policy rule function that accepts a
        dictionary and calls `rule_func(value)` successively, with the
        partial scope set to the key.

        `each` optionally takes a named argument `ref=dict()` to provide
        a built-in lookup for some reference dictionary. If ref is
        provided, `rule_func(ref[key])` is called instead.
        """
        ref_obj = kwargs.get('ref')

        def for_keys(keys):

            @policy_rule_func(m)
            def each_step(key, rule_func):
                if ref_obj is not None:
                    value = ref_obj.get(key)
                    return (
                        regarding(key, unit(value) >> rule_func)
                    )
                return regarding(key, rule_func)

            steps = [
                each_step(key, rule_func)
                for rule_func in rule_funcs
                for key in keys
            ]
            return regarding("", *steps)

        each_rule_func_name = get_call_repr("each", *rule_funcs, **kwargs)
        return policy_rule_func(m, each_rule_func_name)(for_keys)
    return each


each = make_each(List)


#
# Non-Determinism Rules
#

def make_fail(m):
    @policy_rule_func(m)
    def fail():
        def for_partial(partial):
            return m.mzero()
        return for_partial
    return fail


fail = make_fail(List)


def make_match(m):
    @policy_rule_func(m)
    def match(compare_to):
        """
        Given an expected value, selects the currently scoped node and ensures
        it matches expected. If the match results in a new node definition,
        the partial is updated accordingly.

        For non-matches, returns a monadic zero (e.g. if we're building a list
        of policies, this would collapse from [partial] to [])
        """
        def for_partial(partial):
            matches, new_partial = partial.match(compare_to)

            if matches:
                return m.unit((compare_to, new_partial))

            return m.mzero()
        return for_partial
    return match


match = make_match(List)


def make_permit_values(m):
    match = make_match(m)

    @policy_rule_func(m)
    def permit_values(permitted_values):
        """
        Given a list of allowed values, matches the current partial against
        each, forking the non-deterministic computation.
        """
        def for_partial(partial):
            def for_value(value):
                rule = match(value)
                return rule.run(partial)

            monad = m.mzero()
            for value in permitted_values:
                monad = monad.mplus(for_value(value))

            return monad
        return for_partial
    return permit_values


permit_values = make_permit_values(List)


def make_attempt(m):
    mzero = m.mzero
    unit = make_unit(m)
    collect = make_collect(m)

    def attempt(*rules):
        """
        Keeping track of the value and partial it receives,
        if the result of *rules on the partial is mzero,
        then `attempt` returns `unit( (initial_value, initial_policy) )`
        otherwise, `attempt` returns the result of the rules.
        """
        def for_value(value):
            def for_partial(initial_partial):
                op = unit(value) >> collect(*rules)
                result = op.run(initial_partial)

                if result == mzero():
                    return m.unit((value, initial_partial))
                return result
            return for_partial
        attempt_rule_func_name = get_call_repr("attempt", *rules)
        return policy_rule_func(m, attempt_rule_func_name)(for_value)
    return attempt


attempt = make_attempt(List)


def make_catch_attempt(m):
    mzero = m.mzero
    unit = make_unit(m)
    collect = make_collect(m)

    def catch_attempt(catch_rule, *rules):
        """
        Like `attempt`, `catch_attempt` runs a list of policy rule[_func]s,
        but instead of performing no-op on monadic failure, instead runs
        an alternative `catch_rule`
        """
        def for_value(value):
            def for_partial(initial_partial):
                op = unit(value) >> collect(*rules)
                result = op.run(initial_partial)

                if result == mzero():
                    alternative = (unit(value) >> catch_rule).run(
                        initial_partial
                    )
                    return alternative
                return result
            return for_partial
        attempt_rule_func_name = get_call_repr("attempt", catch_rule, *rules)
        return policy_rule_func(m, attempt_rule_func_name)(for_value)
    return catch_attempt


catch_attempt = make_catch_attempt(List)


#
# Context Operators
#

def make_push_context(m):
    append_value = make_append_value(m)

    @policy_rule_func(m)
    def push_context(context):
        """
        Add an additional context to the stack for the partial
        """
        return regarding("/context", append_value(context))
    return push_context


push_context = make_push_context(List)


def make_pop_context(m):
    unit = make_unit(m)

    @policy_rule_func(m)
    def pop_context(passthru):
        """
        Pop the partial's context stack, returning whatever
        value it was called with.
        """
        return regarding("/context", pop_value()) >> unit(passthru)
    return pop_context


pop_context = make_pop_context(List)


def make_wrap_context(m):
    push_context = make_push_context(m)
    pop_context = make_pop_context(m)

    @policy_rule_func(m)
    def wrap_context(context, op):
        """
        Run some operator inside some context
        """
        return push_context(context) >> op >> pop_context

    return wrap_context


wrap_context = make_wrap_context(List)


def make_require_value(m):
    @policy_rule_func(m)
    def require_value(node):
        """
        Returns an mzero (empty list, e.g.) if the provided node
        is missing a value

        :Examples:
        >>> select("/does/not/exist") >> require_value
        []
        """
        def for_partial(partial):
            logger.debug("require_value %r", node)
            if isinstance(node, PolicyNode):
                if not node.value:
                    logger.debug("require_value fail")
                    return m.mzero()
            if not node:
                logger.debug("require_value fail")
                return m.mzero()
            return m.unit((None, partial))
        return for_partial
    return require_value


require_value = make_require_value(List)


def make_forbid_value(m):
    @policy_rule_func(m)
    def forbid_value(node):
        """
        Returns an mzero (empty list, e.g.) if the provided node
        is missing a value

        For instance:
            select("/does/not/exist") >> forbid_value
        returns []
        """
        def for_partial(partial):
            if isinstance(node, PolicyNode):
                if node.value is None:
                    return m.unit((None, partial))
            if node is None:
                return m.unit((None, partial))
            return m.mzero()
        return for_partial
    return forbid_value


forbid_value = make_forbid_value(List)


def make_unless_errors(m):
    policies = make_policies(m)

    @policy_rule_func(m)
    def unless_errors(*rules):
        @policy_rule_func(m)
        def unless_errors_step(rule):
            def for_partial(partial):
                errors = partial.root.get('errors', None)
                if errors:
                    return m.unit((None, partial))
                return rule.run(partial)
            return for_partial
        return policies(*[unless_errors_step(rule) for rule in rules])
    return unless_errors


unless_errors = make_unless_errors(List)


def make_trace(m):
    unit = make_unit(m)
    policies = make_policies(m)

    @policy_rule_func(m)
    def trace(*rule_funcs):
        """
        Collates the current scope, the current node's value,
        and the current policy context and returns it as a dict
        """
        @policy_rule_func(m)
        def trace_step(rule_func):
            def for_partial(partial):
                # collect information
                scope = partial.scope
                value = partial.scope_value
                context_node, _ = partial.select("/context")
                context = context_node.value

                # build obj that gets passed to rule_func
                trace_obj = {
                    "scope": scope,
                    "value": value,
                    "context": context,
                }

                # run rule_func
                if isinstance(rule_func, PolicyRule):
                    rule = rule_func
                else:
                    rule = rule_func(trace_obj)
                results = rule.run(partial)

                # rescope partial for next step
                def for_result(result):
                    value, partial = result
                    _, rescoped_partial = partial.select(
                        scope, set_path=True
                    )
                    return value, rescoped_partial

                return results.fmap(for_result)
            return for_partial

        if not rule_funcs:
            rule_funcs = [unit]

        op = policies(*[trace_step(rule_func) for rule_func in rule_funcs])

        return op
    return trace


trace = make_trace(List)


def make_args_receiver(m):
    unit = make_unit(m)

    def args_receiver(values):
        @policy_rule_func(m)
        def receive(idx, policy_rule):
            def save(value):
                values[idx] = value
                return unit(None)
            return policy_rule >> save
        return receive
    return args_receiver


args_receiver = make_args_receiver(List)



import copy
import logging

from calcifer.contexts import Context
from calcifer.partial import Partial
from calcifer.operators import unless_errors

logger = logging.getLogger(__name__)


class BasePolicy(object):
    ctx_class = Context

    def __init__(self, *args, **kwargs):
        if len(args) == 1 and callable(args[0]):
            method = args[0]
            self(method)

        self.includes = kwargs.get('includes', [])
        self.bind_ref = kwargs.get('bind_ref', False)
        self.args = []

    def __call__(self, method):
        if not hasattr(self, 'method'):
            self.method = method
        return self

    def __get__(self, obj, cls=None):
        self.parent = obj
        return self

    def using(self, *args):
        new_self = copy.deepcopy(self)
        new_self.args = args
        return new_self

    def get_included_policy(self, policy_name):
        return getattr(self.parent, policy_name)

    def pair_included_policy(self, policy):
        if hasattr(self, 'ref'):
            policy.ref = self.ref
        policy.args = self.args

    defaults = {
    }

    def initial_partial(self, obj=None):
        if obj is None:
            obj = {}

        for k, v in self.__class__.defaults.items():
            if k not in obj:
                obj[k] = v

        return Partial.from_obj(obj)

    def run(self, obj):
        new_self = copy.deepcopy(self)

        new_self.ref = obj
        ctx = new_self.context
        policy_rule = ctx.finalize()

        partial = new_self.initial_partial(obj)
        results = [
            new_self.resolve(final) for _, final in policy_rule.run(partial)
        ]

        return results

    def include(self, other):
        new_self = copy.deepcopy(self)
        new_self.includes.append(other)
        return new_self

    @staticmethod
    def resolve(final):
        return final

    @property
    def context(self):
        ctx_class = self.__class__.ctx_class
        ctx = ctx_class(
            name=getattr(self.method, "__name__", None)
        )
        method_args = [ctx]
        if self.bind_ref:
            method_args.append(self.ref)
        method_args += self.args
        self.method(*method_args)
        if getattr(self, 'parent', None):
            # TODO this is a codesmell
            logger.debug("context name: %s", ctx.ctx_name)
            if ctx.ctx_name == 'endpoint_policy':
                ctx.wrapper = lambda policy_rules: unless_errors(*policy_rules)

            includes = copy.copy(self.includes)
            for policy_or_name in includes:
                if isinstance(policy_or_name, str) and hasattr(self, 'parent'):
                    policy = self.get_included_policy(policy_or_name)
                else:
                    policy = policy_or_name
                self.pair_included_policy(policy)  # copy ref and args, e.g.
                ctx.append(policy.context.finalize())
        return ctx


class DefaultPolicy(BasePolicy):
    def resolve(self, final):
        return {k: v for k, v in final.root.items() if k != 'context'}



import copy
import functools
import logging

from calcifer.operators import (
    wrap_context, catch_attempt, trace, collect, unit, policies, regarding,
    fail, args_receiver,
)
from calcifer.monads import (
    PolicyRule, PolicyRuleFunc, get_call_repr,
)

logger = logging.getLogger(__name__)


def ctx_apply(f, ctx_args):
    """
    Creates a promise to call `f` with some values corresponding to
    the contextual values (or regular values) in `ctx_args`
    """
    if not ctx_args:
        return f

    values = []
    missing = set([])
    for idx, arg in enumerate(ctx_args):
        if hasattr(arg, 'finalize'):
            arg = arg.finalize()

        if hasattr(arg, 'bind'):
            values.append(None)
            missing.add((idx, arg))
            continue

        values.append(arg)

    if not missing:
        return f(*values)

    receive = args_receiver(values)

    apply_rule = None
    for idx, policy_rule in missing:
        step = receive(idx, policy_rule)

        if apply_rule is None:
            apply_rule = step
        else:
            apply_rule = apply_rule >> step

    if hasattr(f, 'rule_func'):
        wrapped_func = f.rule_func
    else:
        wrapped_func = f

    @functools.wraps(wrapped_func)
    def finish(_):
        return f(*values)

    return apply_rule >> finish


def wrap_ctx_values(action, args):
    """
    Run some function `action` on args that may be ContextualValues
    (Values provided by a Context's wrapper)
    """
    missing_args = get_missing(args)
    if missing_args:
        return make_incomplete(
            action,
            args,
            missing_args
        )

    return action(args)


def get_missing(args):
    """
    Determine which args are missing obtainable values
    """
    missing_args = set()
    for arg in args:
        if isinstance(arg, ContextualValue):
            missing_args.add(arg)
        if isinstance(arg, Incomplete):
            missing_args.update(arg.missing)

    return missing_args


def make_incomplete(f, args_, missing_args):
    """
    Make a function that takes a dictionary of {ctx_value: true_value}s
    and maps to f(*args) in the right order.
    """
    @functools.wraps(f)
    def complete(true_values):
        args = []
        for arg in args_:
            if isinstance(arg, ContextualValue):
                args.append(true_values[arg])
            elif isinstance(arg, Incomplete):
                got = {
                    ctx_value: true_values[ctx_value]
                    for ctx_value in arg.missing
                }
                args.append(arg.complete(got))
            else:
                args.append(arg)

        return f(tuple(args))

    incomplete = Incomplete(
        func=complete,
        missing=missing_args
    )

    return incomplete


class ContextualValue(object):
    """
    A value that exists in transmission between policy operators.
    Specifically, a value provided by some Context wrapper.

    For instance:

    ctx.each() becomes:
       children() >> each( ... )

    And inside that `each( ... )` is information (field value) that is
    not obtainable elsewhere (due to generic name of `each`)
    """
    def __init__(self, ctx):
        self.ctx = ctx

    def __hash__(self):
        return hash(self.ctx)

    def __eq__(self, other):
        return isinstance(other, ContextualValue) and self.ctx == other.ctx

    def __ne__(self, other):
        return not(self == other)


class Incomplete(object):
    def __init__(self, func, missing, got=None):
        self.missing = missing
        if got is None:
            got = {}
        self.got = got
        self.func = func

    def complete(self, true_values):
        got = copy.copy(self.got)
        missing = copy.copy(self.missing)

        got.update(true_values)
        missing -= set(true_values.keys())

        if not missing:
            return self.func(got)

        return Incomplete(
            func=self.func,
            missing=missing,
            got=got
        )

    def __repr__(self):
        return (
            "Incomplete(missing={})"
        ).format(set([missing.ctx for missing in self.missing]))


class BaseContext(object):
    """
    Underlying implementation of the Context policy builder.

    Contexts comprise two parts:
        1. A wrapper that takes a list of policy rules and returns
           a single policy rule
        2. A list of policy rules and subcontexts

    ctx.finalize() collects this list, finalizes all subcontexts, and wraps
    the resulting list of policy rules into a single policy rule.

    The Context `append` operation is a bit sophisticated:
    A PolicyRuleFunc may be given to append, with a list of values and/or
    _contextual values_, in place of calling the function straightly
    with the values as arguments.

    Contextual values are one of two things: either a policy rule that returns
    some value, or a Context object that finalize()s to a policy rule that
    returns some value. In either case... there's a value we're after, we just
    don't have it yet. So we need to remember everything we do to that value
    until the value exists. (Contextual values can also just be straight
    values)

    Once we have this, we can come very close to writing policies as if they
    were just regular operations on python variables.
    """
    def __init__(self, wrapper=None, *ctx_args, **kwargs):
        self.items = []
        self.ctx_name = kwargs.get('name', None)
        self.error_handler = None

        if wrapper is None:
            wrapper = self.__class__.get_default_wrapper()

        self.wrapper = wrapper
        self.ctx_args = ctx_args

    @staticmethod
    def get_default_wrapper():
        return lambda policy_rules: policies(*policy_rules)

    @staticmethod
    def is_policy_rule(value):
        return isinstance(value, PolicyRule)

    @staticmethod
    def is_policy_rule_func(value):
        return isinstance(value, PolicyRuleFunc)

    @property
    def value(self):
        return ContextualValue(self)

    def append(self, item, *args):
        """
        Append a policy operation to the Context.

        If it's a regular function or a policy rule function, and *args
        are supplied, use ctx_apply to create a "promise" and append
        that instead.
        """
        if not args:
            self.items.append(item)
        else:
            self.items.append(
                wrap_ctx_values(
                    lambda args: ctx_apply(item, args),
                    args
                )
            )
        return self

    def finalize(self):
        """
        Performs all syntactic manipulations to subcontexts and contained
        policy rules and returns a single policy rule aggregate.
        """
        finalized_items, finalized_error_handler = self.get_finalized_items()
        wrapped = self.wrap(finalized_items, finalized_error_handler)

        return wrapped

    def get_finalized_items(self):
        finalized_items = [
            self._finalize_item(item, getattr(self, 'value', None))
            for item in self.items
            if self._warrants_inclusion(item)
        ]

        if not self.error_handler:
            finalized_error_handler = None

        finalized_error_handler = self._finalize_item(
            self.error_handler, getattr(self, 'value', None)
        )

        return finalized_items, finalized_error_handler

    @classmethod
    def _finalize_item(cls, item, provided_ctx_value=None):
        finalize_item = cls._finalize_item

        if hasattr(item, 'finalize'):
            return finalize_item(item.finalize(), provided_ctx_value)
        if isinstance(item, Incomplete) and provided_ctx_value in item.missing:
            def make_appended_func(item, ctx_value):
                @functools.wraps(item.func)
                def appended_func(value):
                    return item.complete({ctx_value: value})
                return appended_func

            def make_incomplete_func(item, ctx_value):
                @functools.wraps(item.func)
                def incomplete_func(true_values):
                    missing_one = item.complete(true_values)
                    return make_appended_func(missing_one, ctx_value)
                return incomplete_func

            missing = item.missing - set([provided_ctx_value])
            if not missing:
                return make_appended_func(item, provided_ctx_value)

            return Incomplete(
                func=make_incomplete_func(item, provided_ctx_value),
                missing=missing
            )
        return item

    def wrap(self, items, error_handler):
        def make_action(ctx_wrapper, ctx_name, num_ctx_args):
            @functools.wraps(ctx_wrapper)
            def action(items):
                error_handler = items[0]
                ctx_args = items[1:num_ctx_args + 1]
                items = items[num_ctx_args + 1:]
                wrapped = ctx_apply(ctx_wrapper(items), ctx_args)

                if ctx_name:
                    if error_handler:
                        ctx_frame = ContextFrame(
                            ctx_name, wrapped.ast, error_handler
                        )
                    else:
                        ctx_frame = ContextFrame(
                            ctx_name, wrapped.ast
                        )

                    wrapped = wrap_context(ctx_frame, wrapped)
                return wrapped
            return action

        action = make_action(
            self.wrapper,
            self.ctx_name,
            len(self.ctx_args)
        )

        wrapped = wrap_ctx_values(
            action,
            (error_handler,) + self.ctx_args + tuple(items)
        )
        return wrapped

    @staticmethod
    def _warrants_inclusion(item):
        """
        Just to note, as an optimization, don't include policies for
        subcontexts that contain no operations
        """
        if not hasattr(item, 'finalize'):
            return True
        subctx = item
        return len(subctx.items) > 0

    def subctx(self, wrapper=None, *ctx_args):
        """
        Creates and returns another Context contained within this one.
        Much like `append`, can be provided *ctx_args that is used
        to convert `wrapper` into a promise for when the `ctx_args` are
        resolved to values.

        `wrapper`'s type is one of two things:
          - Either a function from a list of policy rules to 1 policy rule
          - or, a function(policy_rules) to a function(*ctx_args) that
            returns 1 policy rule
        """
        sub = self.__class__(wrapper, *ctx_args)
        self.append(sub)
        return sub

    def named_subctx(self, name, wrapper=None, *ctx_args):
        sub = self.__class__(wrapper, *ctx_args)
        sub.ctx_name = name
        self.append(sub)
        return sub

    def attempt_catch(self):
        def attempt_wrapper(policy_rules):
            return catch_attempt(policy_rules[0], *policy_rules[1:])

        attempt_ctx = self.subctx(attempt_wrapper)
        catch = attempt_ctx.trace(attempt_ctx.value)

        return attempt_ctx, catch

    def trace(self, value=None):
        def trace_for_policy_rules(policy_rules):
            def trace_for_true_value(true_value):
                return unit(true_value) >> trace(*policy_rules)
            return trace_for_true_value

        return self.subctx(trace_for_policy_rules, value)

    def or_catch(self):
        """
        Rebuild items so that the last thing was actually done in an
        attempt/catch context
        """
        last = self.items.pop()
        attempt_ctx, catch_ctx = self.attempt_catch()
        attempt_ctx.append(last)
        return catch_ctx

    def apply(self, func, *args):
        def apply_for_policy_rules(policy_rules):
            @functools.wraps(func)
            def apply_for_true_args(*true_args):
                return collect(*policy_rules)(func(*true_args))
            return apply_for_true_args

        if hasattr(func, '__name__'):
            func_name = func.__name__
        else:
            func_name = '<anonymousfunction>'

        apply_ctx = self.named_subctx(
            "apply({})".format(func_name),
            apply_for_policy_rules,
            *args
        )
        return apply_ctx

    def memoized_apply(self, func, *args):
        class MemoKey(object):
            def __init__(self, *true_args):
                self.true_args = true_args

            def __lt__(self, other):
                return self.true_args < other.true_args

            def __hash__(self):
                return make_hash(self.true_args)

            def __eq__(self, other):
                return self.true_args == other.true_args

            def __repr__(self):
                return "<MemoKey {}>".format(
                    get_call_repr(*self.true_args)
                )

        @functools.wraps(func)
        def memoized_func(*true_args):
            if not hasattr(func, 'memo'):
                func.memo = {}
            key = MemoKey(*true_args)
            if key in func.memo:
                logger.debug("Found memo key: %r", key)
                return func.memo[key]
            result = func(*true_args)
            logger.debug("Adding memo key: %r", key)
            logger.debug("Existing memo keys: %r", func.memo.keys())
            func.memo[key] = result
            return result

        return self.apply(memoized_func, *args)

    def check(self, func, *func_args):
        def make_check_wrapper(func):
            def check_wrapper(policy_rules):
                @functools.wraps(func)
                def eval_wrapper(*true_func_args):
                    func_result = func(*true_func_args)
                    if func_result:
                        return collect(*policy_rules)(func_result)
                    return policies()
                return eval_wrapper
            return check_wrapper

        ctx_name = "check({})".format(func.__name__)
        subctx = self.named_subctx(
            ctx_name,
            make_check_wrapper(func),
            *func_args
        )
        return subctx

    def scope_item_subctx(self, parent, child, name=None):
        def scope_item_subctx_for_policy_rules(policy_rules):
            def scope_item_subctx_for_true_relations(true_parent, true_child):
                return regarding(
                    "{}/{}".format(true_parent, true_child),
                    *policy_rules
                )
            return scope_item_subctx_for_true_relations
        subctx = self.subctx(scope_item_subctx_for_policy_rules, parent, child)
        if name is not None:
            subctx.ctx_name = name
        return subctx

    def scope_subctx(self, scope, name=None):
        def scope_subctx_for_policy_rules(policy_rules):
            def scope_subctx_for_true_scope(true_scope):
                return regarding(true_scope, *policy_rules)
            return scope_subctx_for_true_scope
        subctx = self.subctx(scope_subctx_for_policy_rules, scope)
        if name is not None:
            subctx.ctx_name = name
        return subctx

    def select(self, scope):
        return self.scope_subctx(scope, 'select("{}")'.format(scope))

    def fail(self):
        self.append(fail())
        return self

    def __repr__(self):
        return (
            "<{} name='{}'>"
        ).format(self.__class__.__name__, self.ctx_name)


class ContextFrame(object):
    def __init__(self, name, ast, error_handler=None):
        self.name = name
        self.policy_ast = ast
        self.error_handler = error_handler

    def __repr__(self):
        return "<policy '{}'>".format(self.name)

    def __deepcopy__(self, memo):
        # you get a new object but you're not copying that AST
        return ContextFrame(self.name, self.policy_ast, self.error_handler)


# from http://stackoverflow.com/questions/5884066/hashing-a-python-dictionary
def make_hash(o):
    """
    Makes a hash from a dictionary, list, tuple or set to any level, that contains
    only other hashable types (including any lists, tuples, sets, and
    dictionaries).
    """
    if isinstance(o, (set, tuple, list)):
        return hash(tuple([make_hash(e) for e in o]))
    elif not isinstance(o, dict):
        return hash(o)

    new_o = copy.deepcopy(o)
    for k, v in new_o.items():
        new_o[k] = make_hash(v)

    return hash(tuple(frozenset(sorted(new_o.items()))))



# pylint: disable=no-self-argument
import unittest
from unittest import TestCase

from calcifer.partial import Partial

from calcifer.policy import BasePolicy


class PolicyProviderTestCase(TestCase):
    def test_decorator(self):
        class HasPolicy(object):
            class Policy(BasePolicy):
                pass

            @Policy
            def a_policy(ctx):
                ctx.select("/sender").require()

        policy_haver = HasPolicy()

        a_policy = policy_haver.a_policy

        self.assertIsInstance(a_policy, HasPolicy.Policy)

        results = a_policy.run({"sender": "someone"})
        self.assertEqual(len(results), 1)
        self.assertIsInstance(results[0], Partial)

    def test_resolve(self):
        class HasPolicy(object):
            class Policy(BasePolicy):
                def resolve(self, policy_final):
                    return policy_final.root['somenode']

            @Policy
            def a_policy(ctx):
                ctx.select("/sender").require()

        policy_haver = HasPolicy()
        a_policy = policy_haver.a_policy
        results = a_policy.run({
            "sender": "someone",
            "somenode": "bajillion"
        })
        self.assertEqual(results[0], "bajillion")

    def test_includes(self):
        class HasPolicy(object):
            class Policy(BasePolicy):
                @staticmethod
                def resolve(final):
                    return final.root['list']

            @Policy(includes=['b', 'c'])
            def a(ctx):
                ctx.select("/list").append_value(1)

            @Policy
            def b(ctx):
                ctx.select("/list").append_value(2)

            @Policy
            def c(ctx):
                ctx.select("/list").append_value(3)

        policy_haver = HasPolicy()
        a_policy = policy_haver.a
        results = a_policy.run({
            "list": []
        })
        self.assertEqual(results[0], [1, 2, 3])


if __name__ == '__main__':
    unittest.main()



from calcifer.partial import Partial


def run_policy(policy_rule, obj=None):
    if obj is None:
        obj = {}

    if 'context' not in obj:
        obj['context'] = []
    if 'errors' not in obj:
        obj['errors'] = []

    partial = Partial.from_obj(obj)
    _, policy_final = policy_rule.run(partial)[0]
    return policy_final.root



from abc import ABCMeta, abstractmethod


class Definition:
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_template(self):
        pass

    @abstractmethod
    def match(self, value):
        pass

    def __repr__(self):
        return "Definition()"


class Value(Definition):
    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    def match(self, value):
        return (self.value == value, self)

    def get_template(self):
        return self.value

    def __repr__(self):
        return (
            "Value(value={})"
        ).format(self.value)

    def __eq__(self, other):
        return (
            isinstance(other, Value) and other.value == self.value
        )


class Field(Definition):
    def __init__(self, field_type=None, **params):
        self._params = params
        if field_type:
            self._params['type'] = field_type
        if 'required' not in params:
            self._params['required'] = False

    @property
    def params(self):
        return self._params

    @property
    def value(self):
        return self._params.get('value')

    def get_template(self):
        return self.params

    def match(self, value):
        if 'value' in self.params:
            return (self.params['value'] == value, self)

        new_params = {k: v for k, v in self.params.items()}
        new_params['value'] = value
        return (True, Field(**new_params))

    def __repr__(self):
        args = ['{}={}'.format(k, v) for k, v in self.params.items()]
        return "Field({})".format(", ".join(args))

    def __eq__(self, other):
        return (
            isinstance(other, Field) and other.params == self.params
        )



"""
HERE THERE BE DRAGONS
"""
from calcifer.tree import UnknownPolicyNode


class Zipper(object):
    def __init__(self, breadcrumbs, node=None):
        """
        Initialize a zipper
        It takes two arguments:
            - breadcrumbs is a list of Breadcrumb objects
            - node is a PolicyNode
        """
        self.breadcrumbs = breadcrumbs
        if node is None:
            node = UnknownPolicyNode()
        self.node = node

    @property
    def root(self):
        new_zipper = self
        for _ in range(len(self.breadcrumbs)):
            new_zipper = new_zipper.up()
        return new_zipper

    @property
    def path(self):
        mapped_crumbs = list(map(lambda breadcrumb: breadcrumb.step_taken, self.breadcrumbs))
        return list(reversed(mapped_crumbs))

    def select(self, path):
        new_zipper = self
        for step in path:
            new_zipper = new_zipper.down(step)
        return new_zipper

    def down(self, step):
        chosen_node, new_node, steps_not_taken = self.node.choose(step)
        new_breadcrumbs = [Breadcrumb(step, new_node, steps_not_taken)] + self.breadcrumbs
        return Zipper(new_breadcrumbs, chosen_node)

    def set_node(self, node):
        return Zipper(self.breadcrumbs, node)

    def up(self):
        first = self.breadcrumbs[0]
        rest = self.breadcrumbs[1:]

        possible_steps = {first.step_taken: self.node}
        possible_steps.update(first.steps_not_taken)

        empty_parent = first.from_node
        new_node = empty_parent.reconstruct(possible_steps)

        return Zipper(rest, new_node)


class Breadcrumb(object):
    def __init__(self, step_taken, from_node, steps_not_taken=None):
        self.step_taken = step_taken
        self.from_node = from_node
        if steps_not_taken is None:
            steps_not_taken = {}
        self.steps_not_taken = steps_not_taken



import unittest
from unittest import TestCase

from pymonad import Just, List, Maybe

from calcifer.monads import (
    Identity, policy_rule_func
)
from calcifer.tree import (
    LeafPolicyNode, DictPolicyNode, UnknownPolicyNode, Value
)
from calcifer import (
    Partial,
    set_value, select, check, policies, regarding, fail, match, attempt,
    permit_values, define_as, children, each, scope,
)
from calcifer import operators


# set up the operators for the Identity and Maybe monads for
# testing
set_valueI = operators.make_set_value(Identity)
policyI = operators.make_policies(Identity)
regardingI = operators.make_regarding(Identity)
selectI = operators.make_select(Identity)
matchI = operators.make_match(Identity)

set_valueM = operators.make_set_value(Maybe)
policyM = operators.make_policies(Maybe)
regardingM = operators.make_regarding(Maybe)
selectM = operators.make_select(Maybe)
matchM = operators.make_match(Maybe)


class PolicyTestCase(TestCase):
    def test_select(self):
        policy = DictPolicyNode()
        foo_node = DictPolicyNode()
        bar_node = LeafPolicyNode(Value(5))

        foo_node['bar'] = bar_node
        policy["foo"] = foo_node

        partial = Partial(policy)

        # select existing partial
        path = ["foo", "bar"]
        scope = "/{}".format("/".join(path))
        node, new_partial = partial.select(scope)
        self.assertEqual(partial.root, new_partial.root)
        self.assertEqual(LeafPolicyNode(Value(5)), node)
        self.assertEqual(path, new_partial.path)

        # select new partial
        path = ["foo", "baz"]
        scope = "/{}".format("/".join(path))
        value, new_partial = partial.select(scope)
        self.assertNotEqual(partial.root, new_partial.root)
        self.assertEqual(UnknownPolicyNode(), value)
        self.assertEqual(path, new_partial.path)

        # make sure /foo/bar still exists in new_partial
        path = ["foo", "bar"]
        scope = "/{}".format("/".join(path))
        value, new_new_partial = new_partial.select(scope)
        self.assertEqual(LeafPolicyNode(Value(5)), value)

    def test_select_no_set_path(self):
        policy = DictPolicyNode()
        foo_node = DictPolicyNode()
        bar_node = LeafPolicyNode(Value(5))
        foo_node["bar"] = bar_node
        policy["foo"] = foo_node

        item = Partial(policy)

        path = "/foo/bar"
        value, new_item = item.select(path, set_path=False)
        self.assertEqual(LeafPolicyNode(Value(5)), value)
        self.assertEqual([], new_item.path)


class PolicyBuilderTestCase(TestCase):
    def test_set_valueI(self):
        rule = policyI(set_valueI(5))

        _, item = rule.run(Partial()).getValue()

        value, _ = item.select("")
        self.assertEqual(LeafPolicyNode(Value(5)), value)

    def test_regardingI(self):
        rule = policyI(
            regardingI("/foo", set_valueI(5))
        )
        _, item = rule.run(Partial()).getValue()
        value, _ = item.select("/foo")
        self.assertEqual(LeafPolicyNode(Value(5)), value)

        rule = policyI(
            regardingI("/fields/foo", set_valueI(5))
        )
        _, item = rule.run(Partial()).getValue()
        value, _ = item.select("/fields/foo")
        self.assertEqual(LeafPolicyNode(Value(5)), value)

    def test_regardingI_multiple(self):
        rule = policyI(
            regardingI("/fields/foo", set_valueI(5), set_valueI(6))
        )
        _, item = rule.run(Partial()).getValue()
        value, _ = item.select("/fields/foo")
        self.assertEqual(LeafPolicyNode(Value(6)), value)

    def test_regardingM(self):
        rule = policyM(
            regardingM("/fields/foo", set_valueM("foo")),
        )

        maybe = rule.run(Partial())
        self.assertTrue(isinstance(maybe, Just))
        _, partial = maybe.getValue()
        foo_node, _ = partial.select("/fields/foo")

        self.assertEqual(LeafPolicyNode(Value("foo")), foo_node)

    def test_policyM_multiple(self):
        rule = policyM(
            regardingM("/fields/foo", set_valueM("foo")),
            regardingM("/fields/foo", set_valueM("bar")),
        )

        maybe = rule.run(Partial())
        self.assertTrue(isinstance(maybe, Just))
        _, partial = maybe.getValue()
        foo_node, _ = partial.select("/fields/foo")

        self.assertEqual(LeafPolicyNode(Value("bar")), foo_node)

    def test_regardingM_multiple(self):
        rule = policyM(
            regardingM("/fields/foo", set_valueM("foo"), set_valueM("bar"))
        )

        maybe = rule.run(Partial())
        self.assertTrue(isinstance(maybe, Just))
        _, partial = maybe.getValue()
        foo_node, _ = partial.select("/fields/foo")

        self.assertEqual(LeafPolicyNode(Value("bar")), foo_node)

    def test_matchM(self):
        rule = policyM(
            regardingM("/fields/foo", set_valueM("foo")),
            regardingM(
                "/fields/bar",
                selectM("/fields/foo") >> (
                    lambda foo: set_valueM(foo.value + "bar")
                )
            ),
            regardingM("/fields/bar", matchM("foobar"))
        )

        maybe = rule.run(Partial())
        self.assertTrue(isinstance(maybe, Just))
        _, partial = maybe.getValue()
        foo_node, _ = partial.select("/fields/foo")
        bar_node, _ = partial.select("/fields/bar")
        self.assertEqual("foo", foo_node.value)
        self.assertEqual("foobar", bar_node.value)

    def test_regarding(self):
        rule = policies(
            regarding("/fields/foo", set_value("foo"))
        )

        ps = rule.run(Partial())
        self.assertTrue(isinstance(ps, List))

        results = ps.getValue()
        self.assertEqual(1, len(results))

        _, partial = results[0]

        foo_node, _ = partial.select("/fields/foo")
        self.assertEqual(LeafPolicyNode(Value("foo")), foo_node)

    def test_match(self):
        rule = policies(
            regarding("/fields/foo", set_value("foo")),
            regarding(
                "/fields/bar",
                select("/fields/foo") >> (
                    lambda foo: set_value(foo.value + "bar")
                )
            ),
            regarding("/fields/bar", match("foobar"))
        )

        ps = rule.run(Partial())
        self.assertTrue(isinstance(ps, List))

        results = ps.getValue()
        self.assertEqual(1, len(results))

    def test_match_invalid(self):
        rule = policies(
            regarding("/fields/foo", set_value("foo")),
            regarding(
                "/fields/bar",
                select("/fields/foo") >> (
                    lambda foo: set_value(foo.value + "bar")
                )
            ),
            regarding("/fields/bar", match("barfu"))
        )

        ps = rule.run(Partial())
        self.assertTrue(isinstance(ps, List))

        results = ps.getValue()
        self.assertEqual(0, len(results))

    def test_permit_values(self):
        # case 1, both values come through
        rule = policies(
            regarding("/fields/foo", permit_values(["foo", "bar"])),
        )

        ps = rule.run(Partial())
        self.assertTrue(isinstance(ps, List))

        results = ps.getValue()
        self.assertEqual(2, len(results))

        values = [r[1].select("/fields/foo")[0].value for r in results]
        self.assertEqual(["foo", "bar"], values)

        # case 2, one value gets filtered through
        rule = policies(
            regarding("/fields/foo", permit_values(["foo", "bar"])),
            regarding("/fields/foo", match("foo"))
        )

        ps = rule.run(Partial())
        self.assertTrue(isinstance(ps, List))

        results = ps.getValue()
        self.assertEqual(1, len(results))

        values = [r[1].select("/fields/foo")[0].value for r in results]
        self.assertEqual(["foo"], values)

        # case 3, this one gets tricky... we can do the match() first!
        rule = policies(
            regarding("/fields/foo", match("foo")),
            regarding("/fields/foo", permit_values(["foo", "bar"]))
        )

        ps = rule.run(Partial())
        self.assertTrue(isinstance(ps, List))

        results = ps.getValue()
        self.assertEqual(1, len(results))

        values = [r[1].select("/fields/foo")[0].value for r in results]
        self.assertEqual(["foo"], values)

    def test_attempt(self):
        rule = policies(
            regarding(
                "/fields",
                regarding(
                    "foo",
                    permit_values(["foo", "bar"]),
                    attempt(
                        match("foo"),
                        set_value("foo_updated")
                    ),
                )
            )
        )

        ps = rule.run(Partial())

        self.assertTrue(isinstance(ps, List))

        results = ps.getValue()
        self.assertEqual(2, len(results))

        values = [r[1].select("/fields/foo")[0].value for r in results]
        self.assertEqual(["foo_updated", "bar"], values)

    def test_fail(self):
        rule = policies(
            regarding(
                "/fields",
                regarding(
                    "foo",
                    permit_values(["foo", "bar"]),
                    fail(),
                )
            )
        )

        ps = rule.run(Partial())

        self.assertTrue(isinstance(ps, List))

        results = ps.getValue()
        self.assertEqual(0, len(results))

    def test_check(self):
        def get_policies(value):
            return regarding(
                "/fields/foo",
                (check(lambda: value) >> set_value)
            )

        ps = get_policies(5).run(Partial())
        results = ps.getValue()
        values = [r[1].select("/fields/foo")[0].value for r in results]
        self.assertEqual([5], values)

        ps = get_policies(1000).run(Partial())
        results = ps.getValue()
        values = [r[1].select("/fields/foo")[0].value for r in results]
        self.assertEqual([1000], values)

    def test_define_as(self):
        definition = Value(5)

        rule = policies(
            regarding(
                "/fields/foo",
                define_as(definition)
            )
        )

        ps = rule.run(Partial())

        self.assertTrue(isinstance(ps, List))

        results = ps.getValue()

        values = [r[1].select("/fields/foo")[0].value for r in results]

        self.assertEqual([5], values)

    def test_regarding_return(self):
        rule = regarding("/foo")
        ps = rule.run(Partial.from_obj({"foo": 5}))

        results = ps.getValue()
        values = [r[0] for r in results]

        self.assertEqual([5], values)

    def test_regarding_scoping(self):
        assertEqual = self.assertEqual

        @policy_rule_func
        def expect_scope(expected="/", msg=None):
            def for_actual(actual):
                def checker():
                    return assertEqual(actual, expected, msg)
                return check(checker)

            return scope() >> for_actual

        rule = regarding(
            "",
            expect_scope("/", 0),
            regarding("a", expect_scope("/a", 1)),
            expect_scope("/", 2),
            regarding(
                "b",
                expect_scope("/b", 3),
                regarding("c", expect_scope("/b/c", 4)),
                expect_scope("/b", 5),
            ),
        )

        rule.run(Partial.from_obj({}))

    def test_children(self):
        rule = children()
        ps = rule.run(Partial.from_obj({"foo": 5}))

        results = ps.getValue()
        values = [r[0] for r in results]

        self.assertEqual([['foo']], values)

    def test_children_list(self):
        rule = children()
        ps = rule.run(Partial.from_obj([9, 4, 7, 1, 1]))

        results = ps.getValue()
        values = [r[0] for r in results]

        self.assertEqual([["0", "1", "2", "3", "4"]], values)

    def test_each(self):
        counter = {
            "num": 0
        }

        def increment_set(_):
            counter['num'] += 1
            num = counter['num']
            return set_value(num)

        rule = children() >> each(increment_set)
        ps = rule.run(Partial.from_obj({"a": 0, "b": 0, "c": 0}))

        results = ps.getValue()
        roots = [r[1].root for r in results]

        self.assertEqual(len(roots), 1)
        root = roots[0]
        self.assertIsInstance(root, dict)

        values = sorted(root.values())

        self.assertEqual(values, [1, 2, 3])

    def test_each_list(self):
        def increment(value):
            return set_value(value + 1)

        rule = children() >> each(increment)
        partial = Partial.from_obj([2, 5, 1])
        ps = rule.run(partial)

        results = ps.getValue()
        roots = [r[1].root for r in results]

        self.assertEqual(len(roots), 1)
        values = roots[0]
        self.assertIsInstance(values, list)

        self.assertEqual(values, [3, 6, 2])

    def test_each_ref(self):
        ref_obj = {"a": 7, "b": 3, "c": -1}
        rule = children() >> each(set_value, ref=ref_obj)

        ps = rule.run(Partial.from_obj({"a": 0, "b": 0, "c": 0}))

        results = ps.getValue()
        roots = [r[1].root for r in results]

        self.assertEqual(len(roots), 1)
        root = roots[0]
        self.assertIsInstance(root, dict)

        self.assertEqual(root, ref_obj)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
import codecs

from setuptools import setup, find_packages

exec(open("calcifer/_version.py").read())

def read_requirements_file(filename):
    """Read pip-formatted requirements from a file."""
    with open(filename, 'r') as f:
        return [line.strip() for line in f.readlines()
                if not line.startswith('#')]

setup(
    name='calcifer',
    description='A Python based policy framework.',
    version=__version__,
    packages=find_packages(exclude=['tests', 'tests.*']),
    include_package_data=True,
    long_description=codecs.open('README.rst', encoding='utf-8').read(),
    install_requires=read_requirements_file('requirements.txt'),
    tests_require=read_requirements_file('requirements-tests.txt'),
    test_suite='nose.collector',
    author='DramaFever',
    author_email='admin@dramafever.com',
    url='https://github.com/DramaFever/calcifer',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'License :: Other/Proprietary License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
    ],
)



"""
`calcifer.monads` module

Mainly this module provides an implementation of the StateT monad transformer.
The StateT monad is used for non-deterministic generation of templates for
commands.

Some background: the State monad models a function:
    runState(initial_state) -> (computation_result_value, new_state)
that runs a stateful operation on some initial state. The usage of the State
monad and not some other mechanism means that stateful operations can be
chained together in a semantically rich way. (For examples of this chaining,
see the .operators module)

StateT describes *non-deterministic* stateful operations - operations that may
only possibly return a new state, or operations that return more than one
possible new state, etc.

For commands in this system, the StateT monad is used to specifically
transform the List monad: clients may be allowed to run commands in more than
one way, and thus, the command policy for a given request may indeed return
any number of templates, including zero. This is realized as:
    runStateT(initial_state) -> [(computation_result_value, new_state)]
"""
from abc import ABCMeta
import copy
import inspect
import logging
from pymonad import Monad, List

from calcifer import asts
from calcifer.asts import get_call_repr  # pylint: disable=unused-import

logger = logging.getLogger(__name__)


def stateT(m):
    class StateT(Monad):
        """
        The StateT monad transformer runs stateful computations over an
        internal monad. This allows non-deterministic policy code to
        generate more than one template at a time (StateT over a List).

        Helpful information may be found at:
        https://en.wikibooks.org/wiki/Haskell/Monad_transformers
        """
        @classmethod
        def unit(cls, value):
            return cls(lambda state: m.unit((value, state)))

        def bind(self, function):
            @StateT
            def newState(state):
                def for_state_result(result):
                    # before state transition
                    value, state = result

                    # after state transition
                    run_func = function(value).run
                    m_new_result = run_func(state)

                    return m_new_result

                run_func = self.run
                return run_func(state) >> for_state_result
            return newState

        @property
        def run(self):
            return self.value

        def fmap(self, function):
            return super(StateT, self).fmap(function)

        def amap(self, function):
            return super(StateT, self).amap(function)

    return StateT


class Identity(Monad):
    """
    The Identity monad is provided here to accompany tests on simple state
    mechanisms that do not exhibit non-deterministic behavior.
    """
    def __init__(self, value):
        self.value = value

    def bind(self, function):
        return function(self.value)

    @classmethod
    def unit(cls, value):
        return Identity(value)

    def fmap(self, function):
        return self.__class__(function(self.value))

    def amap(self, function):
        return super(Identity, self).amap(function)


class BasePolicyRule(object):
    pass


def policyM(m):
    class PolicyRule(BasePolicyRule, stateT(m)):
        def __init__(
                self, for_partial, context=None,
                ast=None,
        ):
            if ast is None:
                ast = getattr(for_partial, 'ast', None)

            if context is not None:
                ast = context.with_result(ast)

            self.ast = ast
            super(PolicyRule, self).__init__(for_partial)

        def __repr__(self):
            if self.ast:
                return "<PolicyRule: {}>".format(repr(self.ast))
            return super(PolicyRule, self).__repr__()

        def __hash__(self):
            return hash(self.value)

        def __deepcopy__(self, memo):
            return copy.copy(self)

        def _bind_policy_rule(self, rule):
            def for_operands(left, right):
                def combined_for_partial(initial_partial):
                    m_results = left.run(initial_partial)

                    def for_m_result(m_result):
                        _, partial = m_result
                        return right.run(partial)
                    return m_results >> for_m_result
                return combined_for_partial

            new_ast = asts.Binding(self.ast, rule.ast)
            return PolicyRule(
                for_operands(self, rule), ast=new_ast
            )

        def bind(self, rule_func):
            if isinstance(rule_func, BasePolicyRule):
                return self._bind_policy_rule(rule_func)

            if not isinstance(rule_func, BasePolicyRuleFunc):
                rule_func = policy_rule_funcM(m)(rule_func)

            try:
                binding = super(PolicyRule, self).bind(rule_func)
            except:
                logger.debug(
                    "error binding `%r` to rule_func `%r`",
                    self, rule_func
                )
                raise

            new_ast = asts.Binding(self.ast, rule_func.ast)

            return PolicyRule(
                binding.value, ast=new_ast
            )

        def __rshift__(self, function):
            """
            The bind operator. `>>` and `bind` are equivalent.

            Note: this overrides Monad.__rshift__ because the inherited
            implementation makes it impossible to perform the
            `_bind_policy_rule` optimization above.
            """
            result = self.bind(function)
            if not isinstance(result, Monad):
                raise TypeError("Operator '>>' must return a Monad instance.")
            return result

    return PolicyRule


class BasePolicyRuleFunc:
    __metaclass__ = ABCMeta


def policy_rule_funcM(m, rule_func_name=None):
    def decorator(rule_func):
        class PolicyRuleFunc(BasePolicyRuleFunc):
            def __init__(self, rule_func, rule_func_name=None):
                if rule_func_name is None:
                    if hasattr(rule_func, 'ast'):
                        rule_func_name = repr(rule_func.ast)
                    else:
                        rule_func_name = rule_func.__name__
                if rule_func_name == '<lambda>':
                    rule_func_name = (
                        '<lambda {}:>'
                    ).format(
                        ", ".join(inspect.getargspec(rule_func).args)  # pylint: disable=deprecated-method
                    )

                if rule_func.__doc__:
                    # this is janky but it works.
                    # if someone goes through the trouble of writing a
                    # docstring for a rule func, it should be accessible
                    # with `help()` and nicely readable.
                    self.__class__.__doc__ = rule_func.__doc__
                    self.__class__.__name__ = (
                        "<PolicyRuleFunc {}>".format(rule_func_name)
                    )

                self.ast = asts.PolicyRuleFunc(rule_func_name)
                self.rule_func = rule_func
                self.rule_func_name = rule_func_name

            def __call__(self, *args, **kwargs):
                for_partial = self.rule_func(*args, **kwargs)
                if isinstance(for_partial, BasePolicyRule):
                    for_partial = for_partial.run

                func_call_ast = asts.PolicyRuleFuncCall(
                    self.ast, args, kwargs
                )
                return policyM(m)(
                    for_partial, context=func_call_ast
                )

            def __repr__(self):
                return "<PolicyRuleFunc {}>".format(self.rule_func_name)

        rule_func = PolicyRuleFunc(rule_func, rule_func_name)
        return rule_func
    return decorator


PolicyRule = BasePolicyRule
PolicyRuleFunc = BasePolicyRuleFunc


# decorators
def policy_rule(*args, **kwargs):
    """
    Decorator to properly wrap a function of type
        partial -> (value, partial')

    ie., PolicyRule value

    May be used with or without parentheses.
    """

    if len(args) == 1 and callable(args[0]):
        return policyM(List)(args[0])
    return policyM(List, *args, **kwargs)


def policy_rule_func(*args, **kwargs):
    """
    Decorator to properly wrap a function of type
        value -> PolicyRule value'

    May be used with or without parentheses.

    :param rule_func_name: Optional, to provide additional semantic information
    about the policy rule function
    """
    if len(args) == 1 and callable(args[0]):
        return policy_rule_funcM(List)(args[0])
    return policy_rule_funcM(List, *args, **kwargs)



"""
Overview and Purpose
====================

Contexts are ordered containers of policy rules, functions that
generate policy rules, and/or sub-contexts. Contexts provide
semantic grouping of policy rules and policy rule generation through
means of deferred value resolution.

The primary goal for Context is to allow the semantic expression of request-
processing policies, so that application-level policy code can be written
with minimal regard for the common underlying non-determinism and
error-propagation behaviors.


General Structure / Scope
-------------------------

Structurally, Context comprises the following parts:
 - An ordered list of items contained in the context
 - A wrapper, within which run the context's contained items
 - Additional semantic annotations such as a context name

A context object may be used for either or both of the following purposes:
 - Contexts may be used to represent a distinct semantic grouping of policy
   rules, as might be expressed by business logic or security requirements.
   This may be granular or broad. This semantic grouping may specify some
   control flow for the contained items, or it may define the way in which
   errors are represented.

   Typically, the semantic meaning of each context is along the lines of
   either: "with regard to some field or parameter", or "with regard to some
   condition being the case".

   Using contexts in this fashion follows the builder pattern - applying
   chained method calls to the context object.

 - Or, the context may represent a value that will exist when policy is run
   on a given request. For instance, `ctx.select("merchant_type")` would
   represent the value for that node.

   A context may be used as this deferred value by passing it as an argument
   to any exposed method on any parent context object.

   Contexts may be used as values in either a context free or a context
   specific manner. The former is more typical, just passing the context to
   some method. Context-specific values can be accessed as a property
   `somectx.value`, and can only be used as arguments for methods in contexts
   descending from the value provider (`somectx`)
"""
from calcifer.contexts.context import (
    Context
)



"""
`calcifer.tree` module

This module implements a non-deterministic nested dictionary (tree).
The tree comprises leaf nodes, dict nodes, and "unknown nodes" -- nodes
which are known to exist but undefined beyond that.

Ultimately, the policy tree contains *definitions*, a higher-level abstraction
on "value": LeafPolicyNode uses the property `definition`, which may compare
to specific values or generate a template for procuring the value.
"""
from abc import ABCMeta, abstractmethod
import logging

from calcifer.definitions import Value

logger = logging.getLogger(__name__)


class PolicyNode:
    """
    Abstract class for node tree.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_template(self):
        """
        Generate the template for the node (recursively)
        """
        pass

    @abstractmethod
    def select(self, path=None):
        """
        Traverse the tree and retrieve a specific node with a given path.
        `select` retrieves existing nodes or populates default nodes based
        on path values.

        Returns a tuple of (selected_node, new_root)
        """
        if not path:
            return (self, self)

    @abstractmethod
    def match(self, value):
        """
        `match` compares a node with a given value, possibly returning an
        altered node in the process. For unknown nodes, this means populating
        the node with a leaf node defined as having that value.

        For nodes with a more complex definition, the behavior of `match`
        defers to the definition of the node.
        """
        return False, self

    @abstractmethod
    def choose(self, step):
        """
        Moves down the given step and returns:
        (the chosen node, the new version of itself (list or dict), and a dict of the steps not taken)
        """
        return (None, None, {})

    @abstractmethod
    def reconstruct(self, possible_steps):
        """
        This method takes in a dictionary of possible steps that could be taken and returns a node object
        """
        raise NotImplementedError

    @staticmethod
    def from_obj(obj):
        """
        To facilitate converting nested dict data structures, the static
        method `from_obj` recursively constructs a PolicyNode tree from
        an object
        """
        if isinstance(obj, PolicyNode):
            return obj
        if isinstance(obj, dict):
            return DictPolicyNode(**obj)
        if isinstance(obj, list):
            return ListPolicyNode(*obj)
        return LeafPolicyNode(Value(obj))


class UnknownPolicyNode(PolicyNode):
    def __init__(self):
        pass

    @property
    def value(self):
        return None

    def reconstruct(self, possible_steps):
        raise TypeError

    def get_template(self):
        return {}

    def choose(self, step):
        if isinstance(step, int):
            new_self = ListPolicyNode()
            steps_not_taken = {k: UnknownPolicyNode() for k in range(step)}
        else:
            new_self = DictPolicyNode()
            steps_not_taken = {}
        return (UnknownPolicyNode(), new_self, steps_not_taken)

    def select(self, path=None):
        if not path:
            return (self, self)

        # recurse
        first = path[0]
        rest = path[1:]

        value, subpolicy = UnknownPolicyNode().select(rest)

        return value, DictPolicyNode(**{first: subpolicy})

    def match(self, value):
        return True, LeafPolicyNode(Value(value))

    def __repr__(self):
        return "UnknownPolicyNode()"

    def __eq__(self, other):
        return isinstance(other, UnknownPolicyNode)


class LeafPolicyNode(PolicyNode):
    def __init__(self, definition=None):
        self._definition = definition

    @property
    def definition(self):
        return self._definition

    @property
    def value(self):
        return self._definition.value

    def reconstruct(self, possible_steps):
        if possible_steps:
            raise TypeError
        return self.__class__(self._definition)

    def get_template(self):
        return self.definition.get_template()

    def choose(self, step):
        raise TypeError("You're at the end dummy!")

    def select(self, path=None):
        if path:
            logger.debug((
                "Attempting to select sub-path %r of %r"
            ), path, self)
            raise Exception(
                "Node cannot be traversed, attempted sub-path: {}".format(path)
            )

        return (self, self)

    def match(self, value):
        matches, new_definition = self.definition.match(value)
        return matches, LeafPolicyNode(new_definition)

    def __repr__(self):
        return (
            "LeafPolicyNode("
            "definition={definition}"
            ")"
        ).format(definition=self.definition)

    def __eq__(self, other):
        return (
            isinstance(other, LeafPolicyNode) and
            other.definition == self.definition
        )


class DictPolicyNode(PolicyNode):
    def __init__(self, **nodes):
        self._nodes = {
            k: PolicyNode.from_obj(v)
            for k, v in nodes.items()
        }

    @property
    def nodes(self):
        return self._nodes

    @property
    def keys(self):
        return self._nodes.keys()

    @property
    def value(self):
        return {
            name: node.value
            for name, node in self.nodes.items()
        }

    def reconstruct(self, possible_steps):
        return DictPolicyNode(**possible_steps)

    def choose(self, step):
        chosen_node = self._nodes.get(step, UnknownPolicyNode())
        new_self = self
        steps_not_taken = {k: v for k, v in self._nodes.items() if k != step}
        return chosen_node, new_self, steps_not_taken

    def get_template(self):
        return {
            k: v.get_template() for k, v in self.nodes.items()
        }

    def select(self, path=None):
        if not path:
            return (self, self)

        first = path[0]
        rest = path[1:]

        node, new_first = self[first].select(rest)
        new_nodes = {k: v for k, v in self.nodes.items()}
        new_nodes[first] = new_first

        return node, DictPolicyNode(**new_nodes)

    def match(self, value):
        return False, self

    def __setitem__(self, key, node):
        self._nodes[key] = node

    def __getitem__(self, key):
        if key not in self._nodes:
            return UnknownPolicyNode()
        return self._nodes[key]

    def __repr__(self):
        args = ['{}={}'.format(k, v) for k, v in self.nodes.items()]
        return "DictPolicyNode({})".format(", ".join(args))

    def __eq__(self, other):
        return (
            isinstance(other, DictPolicyNode) and
            other.nodes == self.nodes
        )


class ListPolicyNode(PolicyNode):
    def __init__(self, *nodes):
        self._nodes = [
            PolicyNode.from_obj(v)
            for v in nodes
        ]

    @property
    def nodes(self):
        return self._nodes

    @property
    def keys(self):
        return [key for key in range(len(self._nodes))]

    @property
    def value(self):
        return [
            node.value
            for node in self.nodes
        ]

    def reconstruct(self, possible_steps):
        if not possible_steps:
            return ListPolicyNode()
        highest_key = sorted(possible_steps.keys(), reverse=True)[0]
        return ListPolicyNode(*[
            possible_steps.get(i, UnknownPolicyNode())
            for i in range(highest_key + 1)
        ])

    def choose(self, step):
        if len(self._nodes) > step:
            # We have the step for sure
            chosen_node = self._nodes[step]
        else:
            # step does not exist yet, must populate list with UnknownPolicyNodes
            chosen_node = UnknownPolicyNode()

        new_self = self
        steps_not_taken = {i: self._nodes[i] for i in range(len(self._nodes)) if i != step}
        return chosen_node, new_self, steps_not_taken

    def get_template(self):
        return [
            v.get_template() for v in self.nodes
        ]

    def select(self, path=None):
        if not path:
            return (self, self)

        first = int(path[0])

        rest = path[1:]

        node, new_first = self[first].select(rest)
        new_nodes = [v for v in self.nodes]
        new_nodes[first] = new_first

        return node, ListPolicyNode(*new_nodes)

    def match(self, value):
        return False, self

    def __setitem__(self, key, node):
        key = int(key)
        sparsity = key - len(self._nodes) + 1
        self._nodes.extend([UnknownPolicyNode()] * sparsity)
        self._nodes[key] = node

    def __getitem__(self, key):
        try:
            key = int(key)
            return self._nodes[int(key)]
        except:
            return UnknownPolicyNode()

    def __repr__(self):
        args = ['{}'.format(v) for v in self.nodes]
        return "ListPolicyNode({})".format(", ".join(args))

    def __eq__(self, other):
        return (
            isinstance(other, ListPolicyNode) and
            other.nodes == self.nodes
        )



"""
`calcifer.partial` module

This module is used to provide the specific data structure used in the
stateful computation of command policy.

The data structure has two parts:
- a root policy node
- a pointer to a "current scope"

Operations are provided on Partial that allow the manipulation of either
the policy tree or the pointer, or both.
"""
from calcifer.tree import (
    PolicyNode, UnknownPolicyNode, LeafPolicyNode
)
from calcifer.zipper import Zipper


class Partial(object):
    def __init__(self, root=None, path=None):
        if root is None:
            root = UnknownPolicyNode()
        if path is None:
            path = []

        self.zipper = Zipper([], root).select(path)

    @staticmethod
    def from_obj(obj):
        return Partial(
            root=PolicyNode.from_obj(obj)
        )

    @property
    def root(self):
        return self.zipper.root.node.value

    @property
    def path(self):
        return self.zipper.path

    @property
    def scope(self):
        return "/{}".format("/".join([str(step) for step in self.zipper.path]))

    @property
    def scope_value(self):
        node, _ = self.select(self.scope, set_path=False)
        return node.value

    def get_template(self):
        return self.zipper.root.node.get_template()

    @staticmethod
    def sub_scope(parent_abs='/', child_rel=''):
        rels = []
        if parent_abs == '/':
            rels.append('')
        else:
            rels += parent_abs.split('/')

        if child_rel:
            rels += child_rel.split('/')

        return '/'.join(rels)

    def select(self, scope, set_path=True):
        """
        Select a node at a given scope, possibly setting the path on a newly returned
        partial.

        Cases:
            - If scope begins with "/", it's an absolute path
            - Otherwise, scope is a relative path, and the existing path should be subscoped
        """
        old_scope = self.scope
        old_path = self.zipper.path

        if not scope:
            scope = old_scope
        elif scope[0] != '/':
            scope = self.sub_scope(old_scope, scope)

        # scope should be absolute
        # convert scope to path

        # remove leading slash
        scope = scope[1:]
        if scope:
            selected_path = scope.split("/")  # to remove leading slash
        else:
            selected_path = []

        def maybe_coerce_to_int(step):
            try:
                return int(step)
            except ValueError:
                return step

        selected_path = [
            maybe_coerce_to_int(step) for step in selected_path
        ]

        new_zipper = self.zipper.root.select(selected_path)

        if set_path:
            new_path = selected_path
        else:
            new_path = old_path

        return new_zipper.node, Partial(new_zipper.root.node, new_path)

    def define_as(self, definition):
        existing_value = self.scope_value
        if existing_value:
            valid, new_definition = definition.match(existing_value)
            if not valid:
                return (None, self)
            definition = new_definition

        new_zipper = self.zipper.set_node(LeafPolicyNode(definition))
        partial = Partial(new_zipper.root.node, self.path)
        return definition, partial

    def set_value(self, value, selector=None):
        partial = self
        if selector is not None:
            _, partial = partial.select(selector)
        new_zipper = partial.zipper.set_node(PolicyNode.from_obj(value))

        return (
            value, Partial(new_zipper.root.node, path=self.path)
        )

    def set_node(self, node):
        new_zipper = self.zipper.set_node(node)

        return (
            node, Partial(new_zipper.root.node, path=self.path)
        )

    def match(self, value):
        node, new_self = self.select("")
        matches, new_node = node.match(value)
        _, new_partial = new_self.set_node(new_node)
        if matches:
            return True, new_partial

        return False, self

    def __repr__(self):
        return "Partial(root={}, path={})".format(self.root, self.path)
